import javax.swing.*;
import javax.swing.Timer;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Cursor;
import java.awt.Graphics2D;
import java.util.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.BasicStroke;
import java.awt.geom.Line2D;

public class MyWorld implements ActionListener {
	public static final double DEAFAUL_REFRESH = 0.003;  // 30 [ms]
	public static final double DEFAUL_DELTA = 0.0001;    // 1 [ms]
	public static int WIDTH = 700;  // in pixels
	public static int HEIGHT = 550; // in pixels
	public static int X_ORIGEN = (int)(WIDTH*0.1); // MyWorld space origin (x,y) will be on
	public static int Y_ORIGEN = (int)(HEIGHT*0.9); // (X_ORIGEN,Y_ORIGEN) of the panel space.
	public static AffineTransform SPACE_TRANSFORM;  // transforms (x,y) in (X,Y) of panel.
	public static AffineTransform SPACE_INVERSE_TRANSFORM; // transforms (X,Y) in (x,y) of my world.
	public static Line2D.Double SCALE_LINE;  // 1 [m] line
	private static double AXES_SCALE = 100.0;  // 1 [m] equals 100 [pixels]

	static {
		SPACE_TRANSFORM = AffineTransform.getTranslateInstance(X_ORIGEN, Y_ORIGEN);
		SPACE_TRANSFORM.scale(AXES_SCALE,-AXES_SCALE);  // it also inverts direction of y-coordinate 
		try {
			SPACE_INVERSE_TRANSFORM = SPACE_TRANSFORM.createInverse();
		}catch (NoninvertibleTransformException e){ 
			System.exit(0);
		}
		SCALE_LINE = new Line2D.Double(300,15, 300+AXES_SCALE,15);  // 1[m] line.
	}
	private double t;              // Simulation time
	private Timer pase;            // Clock to refresh the simulation's view. 
	private double delta_t;        // in seconds
	private double refreshPeriod;  // in seconds

	private Vector2D g;
	private float viscosity;
	private PhysicsLabPanel labPanel;
	private Vector<PhysicsElement> elements;
	private Grafic grafico;
	
	/**
 	* Constructor MyWorld
 	*/
	public MyWorld(){
		t = 0;
		refreshPeriod = 0.04;      // initial value 40 [ms]
		delta_t = 0.001;          // initial value 1 [ms]
		pase = new Timer((int)(refreshPeriod*1000), this);
		g = new Vector2D(0,0);  // start with no gravity
		viscosity = 0.0f;        // no friction
		elements = new Vector<PhysicsElement>();
		labPanel = null;
	}
	/**
 	* Asigna el panel de trabajo
	* @param p  Panel en donde vive MyWorld
 	*/
	public void setPanel (PhysicsLabPanel p) {
		labPanel = p;
	}
	/**
 	* Setea delta t a utilizar en la simulacion
	* @param delta  valor de delta t en segundos
 	*/
	public void setDelta_t(double delta) {
		delta_t = delta;
	}
	/**
 	* Setea el periodo con el que se refrescara la pantalla
	* @param rp  valor en segundos
 	*/
	public void setRefreshPeriod (double rp) {
		refreshPeriod = rp;
		pase.setDelay((int)(refreshPeriod*1000)); // convert from [s] to [ms]
	}
	/**
 	* Asigna la gravedad de MyWorld
	* @param gravity  gravedad en m/s^2
 	*/
	public void setGravity (Vector2D gravity) {
		g = gravity;
	} 
	/**
 	* Remueve un elemento de MyWorld
	*  @param e elemento fisico a remover
 	*/
	public void removeElement(PhysicsElement e) {
	    if (e == null) return;
	    elements.remove(e);
	    labPanel.repaint();
	}
	/**
 	* Retorna el valor de la gravedad actual
	* @return g valor de la gravedad
 	*/
	public Vector2D getGravity(){
		return g;
	}
	/**
 	* Asigna la viscosidad aplicada
	* @param v viscosidad en Kg/(sm)
 	*/
	public void setViscosity(float v){
		viscosity = v;
	}
	/**
 	* Retorna la viscosidad actual
	* @return v viscosidad en Kg/(sm)
 	*/
	public float getViscosity(){
		return viscosity;
	}

	/**
 	* Agrega un elemento a MyWorld y lo muestra
	* @param e  elemento a agregar a MyWorld
 	*/
	public void addElement(PhysicsElement e) {
		elements.add(e);
		if (labPanel!=null)
			labPanel.repaint();
	}
	/**
 	* Comienza la simulacion
 	*/
	public void start() {
		if(pase.isRunning()) return;
		labPanel.desableMouseListener();
		pase.start();      
	}
	/**
 	* Detiene la simulacion
 	*/
	public void stop(){
		pase.stop();
		labPanel.enableMouseListener();
	}
	/** 
   	*  Action Performed 
   	*  Se activa segun refreshPeriod para actualizar el estado de todos los elementos, y el calculo de energia del sistema
 	*  @param event evento que invoca este metodo
 	*/
	public void actionPerformed (ActionEvent event) {  // simulate time passing
		double nextStop=t+refreshPeriod;
		double Ek=0,Ep=0;
		double hmin=100;
		GBall bola;
		GWall muro;
		int orient=0;
		while (t<nextStop) {
			Ek=0;Ep=0;
			for (PhysicsElement e: elements){  	        // for each element
				e.computeNextState(delta_t, this);      // compute next state
				if (e instanceof GWall){		// Si es un Muro
					muro=(GWall)e;
					orient=muro.getOrientation();			
					if(orient==0)			// Si es un Muro Horizontal
									// Define como "Piso" el muro horizontal de posicion mas baja		
						if(muro.getY()< hmin) hmin=muro.getY();				
				}
			}	
			for (PhysicsElement e: elements){       // for all elements
				e.updateState();                // update its state
				if (e instanceof GBall )
         		{
					bola=(GBall)e;
					// Sumatoria de Energia Cinetica de todas las bolas
					Ek+= (bola.getMass()/2)*(bola.getSpeed().moduleSquared());
					// sumatoria de Energia Potencial de todas las bolas
					Ep+= (bola.getMass()*(hmin-bola.getY())*g.getY());
         		}   
			}
			t+=delta_t;
		}
		// Imprime en grafico, el calculo de las energias obtenidas
		grafico.add(Ek,Ep,(float)t,(float)refreshPeriod); 
  		labPanel.repaint();		
	}
	/**
 	* Retorna un objeto cercano a anchor
	* @param anchor  objeto al que se consulta cercania
	* @return e elemento cercano a anchor
 	*/
	public PhysicsElement getNearCollision (PhysicsElement anchor) { // first implementation
		for (PhysicsElement e: elements) 
			if ((anchor != e) && anchor.collide(e))
				return e;
		return null;      
	}
	/**
 	* Retorna el objeto mas cercano a anchor
	* @param anchor  objeto al que se consulta cercania
	* @return e elemento cercano a anchor
 	*/
	public PhysicsElement getNearestCollision (PhysicsElement anchor) {
		double minDistance=Double.MAX_VALUE;
		double distance;
		PhysicsElement nearest=null;
		for (PhysicsElement e: elements) {
			if (anchor!=e) {
				distance = anchor.getVectorBetweenCenters(e).module()
				-anchor.getRadius()-e.getRadius();
				if ((distance< minDistance) && anchor.collide(e)){
					minDistance = distance;
					nearest = e;
				}
			}
		}
		return nearest;      
	}
	/**
 	* Dibuja los componentes de MyWorld en g2d
	* @param g2d espacio donde se grafican los elementos
 	*/
	public void paintComponent(Graphics2D g2d){
		g2d.drawString("ELO329 1er.Sem. 2012,   1 [m] = ", 85, 15);
		g2d.draw(SCALE_LINE);
		g2d.transform(SPACE_TRANSFORM);
		g2d.setStroke(new BasicStroke(0.02f));
		for (PhysicsElement e: elements)
			if (e instanceof Drawable)
				((Drawable)e).draw(g2d);
	}
	/**
 	* Retorna un objeto del tipo Drawable si es que encuentra alguno en las coordenadas (d,f)
	* @param d coordenada x donde buscara algun objeto
	* @param f coordenada x donde buscara algun objeto
	* @return de objeto encontrado en la coordenada, de no encontrarse ningun objeto retorna null
 	*/
	public Drawable find(double d, double f) {
		for (PhysicsElement e: elements)
			if (e instanceof Drawable){
				Drawable de = (Drawable) e;
				if (de.contains(d,f)) return de;
			}
		return null;
	}  
	/**
 	* Configura el cursor como una cruz
 	*/
	public void setCursor(){
		labPanel.setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
	} 
	/**
 	* Configura el cursor a la visualizacion normal
 	*/
	public void resetCursor(){
		labPanel.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}
	/**
 	* Genera los graficos de energia 
 	*/
	public void Grafica(JPanel Panel, double plotWindowTime){ //Genera gráficos
		   grafico = new Grafic("Grafico",Panel,plotWindowTime) ;
	}
	/**
 	* Borra los graficos
 	*/
	public void Grafica_reset(){
		   grafico.borraGrafic();
	}
} 
